#!/usr/bin/python
#
# This file is a portion of the Red Hat Update Agent
# Copyright (c) 1999 - 2002 Red Hat, Inc.  Distributed under GPL
#
# Authors:
#       Cristian Gafton <gafton@redhat.com>
#       Adrian Likins   <alikins@redhat.com>
#
# $Id: sourcesConfig.py,v 1.1 2003/08/11 02:48:53 alikins Exp $

import os
import sys
import string

import up2dateLog
import config

from rhpl.translate import _, N_

# The format for sources v1 is stupid. each entry can only be one line
# each different source type has different info (aieee!) # comment stuff out (duh)

#grab an instance of the log class shared
log = up2dateLog.initLog()

#grab an instance of the config class shared
cfg = config.initUp2dateConfig()


class SourcesConfigFile:
    "class for parsing out the up2date/apt/yum src repo info"
    def __init__(self, filename = None):
        self.repos = []
        self.fileName = filename
        if self.fileName:
            self.load()

    def load(self, filename = None):
        if filename:
            self.fileName = filename
        if not self.fileName:
            return

        if not os.access(self.fileName, os.R_OK):
            print "warning: can't access %s" % self.fileName
            return

        f = open(self.fileName, "r")

	for line in f.readlines():
            # strip comments
            if '#' in line:
                line = line[:string.find(line, '#')]

            line = string.strip(line)
            if not line:
                continue

            data = string.split(line)
            repoType = data[0]
            if data[0] == "up2date":
                self.parseUp2date(line)
            if data[0] == "yum":
                self.parseYum(line)
            if data[0] == "apt":
                self.parseApt(line)
            if data[0] == "dir":
                self.parseDir(line)
#        print self.repos
        
    def parseUp2date(self,line):
        (tmp, url) = string.split(line)
        if url == "default":
            self.repos.append({'type':'up2date', 'url':cfg['serverURL']})
        else:
            self.repos.append({'type':'up2date', 'url':url})

    def parseDir(self, line):
        (tmp, name, path) = string.split(line)
        self.repos.append({'type':'dir','path':path, 'label':name})

    def parseYum(self, line):
        (tmp, name, url) = string.split(line)
        self.repos.append({'type':'yum', 'url':url, 'label':name})

    def parseApt(self, line):
        # of course, the debian one had to be weird
        # atm, we only support http one's
        data = string.split(line)
        name = data[1]
        server = data[2]
        path = data[3]
        dists = data[4:]
        # if multiple dists are appended, make them seperate
        # channels
        for dist in dists:
            self.repos.append({'type':'apt',
                          'url':'%s/%s' % (server, path),
                          'label': "%s-%s" % (name,dist),
                          'dist': dist})

global sources
sources = None
def getSources():
    global sources
    if sources:
        return sources
    scfg = SourcesConfigFile(filename="/etc/sysconfig/rhn/sources")
    sources = scfg.repos
    return sources
    

